@@ -55,7 +55,21 @@
  * Since: 0.11.12
  */
 
+/**
+ * TpAccountChannelRequestDelegatedChannelCb:
+ * @request: a #TpAccountChannelRequest instance
+ * @channel: a #TpChannel
+ * @user_data: arbitrary user-supplied data passed to
+ * tp_account_channel_request_set_delegated_channel_callback()
+ *
+ * Called when a client asked us to delegate @channel to another Handler.
+ * When this function is called you are no longer handling @channel.
+ *
+ * Since: 0.15.3
+ */
+
 #include "telepathy-glib/account-channel-request.h"
+#include "telepathy-glib/account-channel-request-internal.h"
 
 #include <telepathy-glib/automatic-proxy-factory.h>
 #include "telepathy-glib/base-client-internal.h"
@@ -69,6 +83,8 @@
 
 #define DEBUG_FLAG TP_DEBUG_CLIENT
 #include "telepathy-glib/debug-internal.h"
+#include "telepathy-glib/deprecated-internal.h"
+#include "telepathy-glib/simple-client-factory-internal.h"
 #include "telepathy-glib/_gen/signals-marshal.h"
 
 struct _TpAccountChannelRequestClass {
@@ -132,6 +148,10 @@ struct _TpAccountChannelRequestPrivate
   gboolean requested;
 
   ActionType action_type;
+
+  TpAccountChannelRequestDelegatedChannelCb delegated_channel_cb;
+  gpointer delegated_channel_data;
+  GDestroyNotify delegated_channel_destroy;
 };
 
 static void
@@ -140,9 +160,6 @@ tp_account_channel_request_init (TpAccountChannelRequest *self)
   self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
       TP_TYPE_ACCOUNT_CHANNEL_REQUEST,
       TpAccountChannelRequestPrivate);
-
-  self->priv->factory = TP_CLIENT_CHANNEL_FACTORY (
-      tp_automatic_proxy_factory_dup ());
 }
 
 static void
@@ -187,6 +204,13 @@ tp_account_channel_request_dispose (GObject *object)
   tp_clear_object (&self->priv->factory);
   tp_clear_pointer (&self->priv->hints, g_hash_table_unref);
 
+  if (self->priv->delegated_channel_destroy != NULL)
+    {
+      self->priv->delegated_channel_destroy (
+          self->priv->delegated_channel_data);
+      self->priv->delegated_channel_destroy = NULL;
+    }
+
   if (dispose != NULL)
     dispose (object);
 }
@@ -698,8 +722,12 @@ acr_channel_request_succeeded_with_channel (TpChannelRequest *chan_req,
         }
 
       /* Operation will be complete once the channel have been prepared */
-      features = tp_client_channel_factory_dup_channel_features (
-          self->priv->factory, self->priv->channel);
+      if (self->priv->factory != NULL)
+        features = tp_client_channel_factory_dup_channel_features (
+            self->priv->factory, self->priv->channel);
+      else
+        features = tp_simple_client_factory_dup_channel_features (
+            tp_proxy_get_factory (self->priv->account), self->priv->channel);
       g_assert (features != NULL);
 
       tp_proxy_prepare_async (self->priv->channel, (GQuark *) features->data,
@@ -769,16 +797,16 @@ acr_request_cb (TpChannelDispatcher *cd,
 
   DEBUG ("Got ChannelRequest: %s", channel_request_path);
 
-  self->priv->chan_request = tp_channel_request_new (
-      self->priv->dbus, channel_request_path, NULL, &err);
-
+  self->priv->chan_request = _tp_simple_client_factory_ensure_channel_request (
+      tp_proxy_get_factory (self->priv->account), channel_request_path, NULL,
+      &err);
   if (self->priv->chan_request == NULL)
     {
       DEBUG ("Failed to create ChannelRequest: %s", err->message);
       goto fail;
     }
 
-  tp_channel_request_set_channel_factory (self->priv->chan_request,
+  _tp_channel_request_set_channel_factory (self->priv->chan_request,
       self->priv->factory);
 
   self->priv->invalidated_sig = g_signal_connect (self->priv->chan_request,
@@ -810,6 +838,27 @@ fail:
   g_error_free (err);
 }
 
+static void
+delegated_channels_cb (TpBaseClient *client,
+    GPtrArray *channels,
+    gpointer user_data)
+{
+  TpAccountChannelRequest *self = user_data;
+  TpChannel *channel;
+
+  g_return_if_fail (channels->len == 1);
+
+  /* TpBaseClient is supposed to check we are actually handling the channel
+   * before calling this callback so we can assert that's the right one. */
+  channel = g_ptr_array_index (channels, 0);
+  g_return_if_fail (TP_IS_CHANNEL (channel));
+  g_return_if_fail (!tp_strdiff (tp_proxy_get_object_path (channel),
+      tp_proxy_get_object_path (self->priv->channel)));
+
+  self->priv->delegated_channel_cb (self, channel,
+      self->priv->delegated_channel_data);
+}
+
 static void
 request_and_handle_channel_async (TpAccountChannelRequest *self,
     GCancellable *cancellable,
@@ -839,14 +888,21 @@ request_and_handle_channel_async (TpAccountChannelRequest *self,
   self->priv->ensure = ensure;
 
   /* Create a temp handler */
-  self->priv->handler = tp_simple_handler_new (self->priv->dbus, TRUE, FALSE,
+  self->priv->handler = tp_simple_handler_new_with_factory (
+      tp_proxy_get_factory (self->priv->account), TRUE, FALSE,
       "TpGLibRequestAndHandle", TRUE, handle_channels, self, NULL);
   _tp_base_client_set_only_for_account (self->priv->handler,
       self->priv->account);
 
-  tp_base_client_set_channel_factory (self->priv->handler,
+  _tp_base_client_set_channel_factory (self->priv->handler,
       self->priv->factory);
 
+  if (self->priv->delegated_channel_cb != NULL)
+    {
+      tp_base_client_set_delegated_channels_callback (self->priv->handler,
+          delegated_channels_cb, self, NULL);
+    }
+
   if (!tp_base_client_register (self->priv->handler, &error))
     {
       DEBUG ("Failed to register temp handler: %s", error->message);
@@ -1327,15 +1383,26 @@ tp_account_channel_request_ensure_channel_finish (
  * channel.
  *
  * Since: 0.13.2
+ * Deprecated: since 0.15.5. The factory is taken from
+ *  #TpAccountChannelRequest:account.
  */
 void
 tp_account_channel_request_set_channel_factory (TpAccountChannelRequest *self,
     TpClientChannelFactory *factory)
+{
+  _tp_account_channel_request_set_channel_factory (self, factory);
+}
+
+void
+_tp_account_channel_request_set_channel_factory (TpAccountChannelRequest *self,
+    TpClientChannelFactory *factory)
 {
   g_return_if_fail (!self->priv->requested);
 
   tp_clear_object (&self->priv->factory);
-  self->priv->factory = g_object_ref (factory);
+
+  if (factory != NULL)
+    self->priv->factory = g_object_ref (factory);
 }
 
 
@@ -1573,3 +1640,84 @@ tp_account_channel_request_set_hints (TpAccountChannelRequest *self,
   tp_clear_pointer (&self->priv->hints, g_hash_table_unref);
   self->priv->hints = g_hash_table_ref (hints);
 }
+
+/**
+ * tp_account_channel_request_set_delegate_to_preferred_handler:
+ * @self: a #TpAccountChannelRequest
+ * @delegate: %TRUE to request to delegate channels
+ *
+ * If @delegate is %TRUE, asks to the client currently handling the channels to
+ * delegate them to the preferred handler (passed when calling
+ * tp_account_channel_request_ensure_channel_async() for example).
+ *
+ * This function can't be called once @self has been used to request a
+ * channel.
+ *
+ * Since: 0.15.3
+ */
+void
+tp_account_channel_request_set_delegate_to_preferred_handler (
+    TpAccountChannelRequest *self,
+    gboolean delegate)
+{
+  g_return_if_fail (!self->priv->requested);
+
+  if (self->priv->hints == NULL)
+    self->priv->hints = tp_asv_new (NULL, NULL);
+
+  tp_asv_set_boolean (self->priv->hints,
+      "org.freedesktop.Telepathy.ChannelRequest.DelegateToPreferredHandler",
+      delegate);
+}
+
+/**
+ * tp_account_channel_request_set_delegated_channel_callback:
+ * @self: a #TpAccountChannelRequest
+ * @callback: function called the channel requested using @self is
+ * delegated, may not be %NULL
+ * @user_data: arbitrary user-supplied data passed to @callback
+ * @destroy: called with the @user_data as argument, when @self is destroyed
+ *
+ * Turn on support for
+ * the org.freedesktop.Telepathy.ChannelRequest.DelegateToPreferredHandler
+ * hint.
+ *
+ * When receiving a request containing this hint, @self will automatically
+ * delegate the channel to the preferred handler of the request and then call
+ * @callback to inform the client that it is no longer handling this channel.
+ *
+ * @callback may be called any time after (and only after) requesting and
+ * handling the channel (i.e. you have called create_and_handle or
+ * ensure_and_handle).
+ *
+ * This function can't be called once @self has been used to request a
+ * channel.
+ *
+ * See also: tp_base_client_set_delegated_channels_callback()
+ *
+ * Since: 0.15.3
+ */
+void
+tp_account_channel_request_set_delegated_channel_callback (
+    TpAccountChannelRequest *self,
+    TpAccountChannelRequestDelegatedChannelCb callback,
+    gpointer user_data,
+    GDestroyNotify destroy)
+{
+  g_return_if_fail (TP_IS_ACCOUNT_CHANNEL_REQUEST (self));
+  g_return_if_fail (!self->priv->requested);
+
+  g_return_if_fail (self->priv->delegated_channel_cb == NULL);
+
+  self->priv->delegated_channel_cb = callback;
+  self->priv->delegated_channel_data = user_data;
+  self->priv->delegated_channel_destroy = destroy;
+}
+
+TpBaseClient *
+_tp_account_channel_request_get_client (TpAccountChannelRequest *self)
+{
+  g_return_val_if_fail (TP_IS_ACCOUNT_CHANNEL_REQUEST (self), NULL);
+
+  return self->priv->handler;
+}
