@@ -23,8 +23,9 @@
 #define TP_ACCOUNT_MANAGER_H
 
 #include <telepathy-glib/account.h>
-#include <telepathy-glib/proxy.h>
 #include <telepathy-glib/dbus.h>
+#include <telepathy-glib/defs.h>
+#include <telepathy-glib/proxy.h>
 
 G_BEGIN_DECLS
 
@@ -72,12 +73,17 @@ GQuark tp_account_manager_get_feature_quark_core (void) G_GNUC_CONST;
 TpAccountManager *tp_account_manager_new (TpDBusDaemon *bus_daemon)
   G_GNUC_WARN_UNUSED_RESULT;
 
+TpAccountManager *tp_account_manager_new_with_factory (
+    TpSimpleClientFactory *factory) G_GNUC_WARN_UNUSED_RESULT;
+
+void tp_account_manager_set_default (TpAccountManager *manager);
 TpAccountManager *tp_account_manager_dup (void) G_GNUC_WARN_UNUSED_RESULT;
 
 void tp_account_manager_init_known_interfaces (void);
 
 TpAccount *tp_account_manager_ensure_account (TpAccountManager *manager,
-    const gchar *path);
+    const gchar *path)
+    _TP_GNUC_DEPRECATED_FOR (tp_simple_client_factory_ensure_account);
 
 GList *tp_account_manager_get_valid_accounts (TpAccountManager *manager)
   G_GNUC_WARN_UNUSED_RESULT;
