@@ -22,6 +22,7 @@
 #include "telepathy-glib/channel-dispatch-operation.h"
 #include "telepathy-glib/channel-dispatch-operation-internal.h"
 
+#include <telepathy-glib/base-client-internal.h>
 #include <telepathy-glib/channel.h>
 #include <telepathy-glib/defs.h>
 #include <telepathy-glib/errors.h>
@@ -30,10 +31,12 @@
 #include <telepathy-glib/proxy-internal.h>
 #include <telepathy-glib/proxy-subclass.h>
 #include <telepathy-glib/util.h>
+#include <telepathy-glib/util-internal.h>
 
 #define DEBUG_FLAG TP_DEBUG_DISPATCHER
 #include "telepathy-glib/dbus-internal.h"
 #include "telepathy-glib/debug-internal.h"
+#include "telepathy-glib/simple-client-factory-internal.h"
 #include "telepathy-glib/_gen/signals-marshal.h"
 
 #include "telepathy-glib/_gen/tp-cli-channel-dispatch-operation-body.h"
@@ -114,8 +117,6 @@ struct _TpChannelDispatchOperationPrivate {
   GPtrArray *channels;
   GStrv possible_handlers;
   GHashTable *immutable_properties;
-
-  gboolean preparing_core;
 };
 
 enum
@@ -243,7 +244,6 @@ static void
 maybe_set_connection (TpChannelDispatchOperation *self,
     const gchar *path)
 {
-  TpDBusDaemon *dbus;
   GError *error = NULL;
 
   if (self->priv->connection != NULL)
@@ -252,9 +252,8 @@ maybe_set_connection (TpChannelDispatchOperation *self,
   if (path == NULL)
     return;
 
-  dbus = tp_proxy_get_dbus_daemon (self);
-
-  self->priv->connection = tp_connection_new (dbus, NULL, path, &error);
+  self->priv->connection = tp_simple_client_factory_ensure_connection (
+      tp_proxy_get_factory (self), path, NULL, &error);
   if (self->priv->connection == NULL)
     {
       DEBUG ("Failed to create connection %s: %s", path, error->message);
@@ -277,7 +276,6 @@ static void
 maybe_set_account (TpChannelDispatchOperation *self,
     const gchar *path)
 {
-  TpDBusDaemon *dbus;
   GError *error = NULL;
 
   if (self->priv->account != NULL)
@@ -286,9 +284,8 @@ maybe_set_account (TpChannelDispatchOperation *self,
   if (path == NULL)
     return;
 
-  dbus = tp_proxy_get_dbus_daemon (self);
-
-  self->priv->account = tp_account_new (dbus, path, &error);
+  self->priv->account = tp_simple_client_factory_ensure_account (
+      tp_proxy_get_factory (self), path, NULL, &error);
   if (self->priv->account == NULL)
     {
       DEBUG ("Failed to create account %s: %s", path, error->message);
@@ -334,25 +331,10 @@ static void
 maybe_set_interfaces (TpChannelDispatchOperation *self,
     const gchar **interfaces)
 {
-  const gchar **iter;
-
   if (interfaces == NULL)
     return;
 
-  for (iter = interfaces; *iter != NULL; iter++)
-    {
-      DEBUG ("- %s", *iter);
-
-      if (tp_dbus_check_valid_interface_name (*iter, NULL))
-        {
-          GQuark q = g_quark_from_string (*iter);
-          tp_proxy_add_interface_by_id ((TpProxy *) self, q);
-        }
-      else
-        {
-          DEBUG ("\tInterface %s not valid, ignoring it", *iter);
-        }
-    }
+  tp_proxy_add_interfaces ((TpProxy *) self, interfaces);
 
   g_hash_table_insert (self->priv->immutable_properties,
       g_strdup (TP_PROP_CHANNEL_DISPATCH_OPERATION_INTERFACES),
@@ -380,29 +362,9 @@ tp_channel_dispatch_operation_set_property (GObject *object,
         break;
 
       case PROP_CHANNELS:
-        {
-          GPtrArray *tmp;
-          guint i;
-
-          g_assert (self->priv->channels == NULL);  /* construct-only */
-
-          /* g_value_dup_boxed returns a new reference to the same
-           * GPtrArray which is not what we want (removing a channel from the
-           * CDO array shouldn't remove it from the caller). Copying the
-           * GPtrArray to avoid this problem.*/
-          tmp = g_value_get_boxed (value);
-
-          if (tmp == NULL)
-            break;
-
-          self->priv->channels = g_ptr_array_sized_new (tmp->len);
-          g_ptr_array_set_free_func (self->priv->channels,
-              (GDestroyNotify) g_object_unref);
-
-          for (i = 0; i < tmp->len; i++)
-            g_ptr_array_add (self->priv->channels,
-                g_object_ref (g_ptr_array_index (tmp, i)));
-        }
+        g_assert (self->priv->channels == NULL);  /* construct-only */
+        _tp_channel_dispatch_operation_ensure_channels (self,
+            g_value_get_boxed (value));
         break;
 
       case PROP_CDO_PROPERTIES:
@@ -438,6 +400,8 @@ tp_channel_dispatch_operation_constructed (GObject *object)
 
   g_return_if_fail (tp_proxy_get_dbus_daemon (self) != NULL);
 
+  _tp_proxy_ensure_factory (self, NULL);
+
   maybe_set_connection (self,
       tp_asv_get_boxed (self->priv->immutable_properties,
         TP_PROP_CHANNEL_DISPATCH_OPERATION_CONNECTION,
@@ -551,9 +515,8 @@ update_channels_array (TpChannelDispatchOperation *self,
       old = self->priv->channels;
     }
 
-  self->priv->channels = g_ptr_array_sized_new (channels->len);
-  g_ptr_array_set_free_func (self->priv->channels,
-      (GDestroyNotify) g_object_unref);
+  self->priv->channels = _tp_g_ptr_array_new_full (channels->len,
+      g_object_unref);
 
   for (i = 0; i < channels->len; i++)
     {
@@ -574,7 +537,8 @@ update_channels_array (TpChannelDispatchOperation *self,
         }
       else
         {
-          channel = tp_channel_new_from_properties (self->priv->connection,
+          channel = tp_simple_client_factory_ensure_channel (
+              tp_proxy_get_factory (self), self->priv->connection,
               path, chan_props, &err);
 
           if (channel == NULL)
@@ -602,12 +566,11 @@ get_dispatch_operation_prop_cb (TpProxy *proxy,
     GObject *weak_object)
 {
   TpChannelDispatchOperation *self = (TpChannelDispatchOperation *) proxy;
+  GSimpleAsyncResult *result = user_data;
   gboolean prepared = TRUE;
   GPtrArray *channels;
   GError *e = NULL;
 
-  self->priv->preparing_core = FALSE;
-
   if (error != NULL)
     {
       DEBUG ("Failed to fetch ChannelDispatchOperation properties: %s",
@@ -682,8 +645,10 @@ get_dispatch_operation_prop_cb (TpProxy *proxy,
   g_object_notify ((GObject *) self, "cdo-properties");
 
 out:
-  _tp_proxy_set_feature_prepared (proxy,
-      TP_CHANNEL_DISPATCH_OPERATION_FEATURE_CORE, prepared);
+  if (e != NULL)
+    g_simple_async_result_set_from_error (result, e);
+
+  g_simple_async_result_complete (result);
 
   if (!prepared)
     {
@@ -693,26 +658,21 @@ out:
 }
 
 static void
-maybe_prepare_core (TpProxy *proxy)
+prepare_core_async (TpProxy *proxy,
+    const TpProxyFeature *feature,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
 {
   TpChannelDispatchOperation *self = (TpChannelDispatchOperation *) proxy;
+  GSimpleAsyncResult *result;
 
-  if (tp_proxy_is_prepared (proxy, TP_CHANNEL_DISPATCH_OPERATION_FEATURE_CORE))
-    return;   /* already done */
-
-  if (self->priv->preparing_core)
-    return;   /* already running */
-
-  if (!_tp_proxy_is_preparing (proxy,
-        TP_CHANNEL_DISPATCH_OPERATION_FEATURE_CORE))
-    return;   /* not interested right now */
-
-  self->priv->preparing_core = TRUE;
+  result = g_simple_async_result_new ((GObject *) proxy, callback, user_data,
+      prepare_core_async);
 
   tp_cli_dbus_properties_call_get_all (self, -1,
       TP_IFACE_CHANNEL_DISPATCH_OPERATION,
       get_dispatch_operation_prop_cb,
-      NULL, NULL, NULL);
+      result, g_object_unref, NULL);
 }
 
 enum {
@@ -730,7 +690,7 @@ tp_channel_dispatch_operation_list_features (TpProxyClass *cls G_GNUC_UNUSED)
 
   features[FEAT_CORE].name = TP_CHANNEL_DISPATCH_OPERATION_FEATURE_CORE;
   features[FEAT_CORE].core = TRUE;
-  features[FEAT_CORE].start_preparing = maybe_prepare_core;
+  features[FEAT_CORE].prepare_async = prepare_core_async;
 
   /* assert that the terminator at the end is there */
   g_assert (features[N_FEAT].name == 0);
@@ -932,6 +892,17 @@ tp_channel_dispatch_operation_new (TpDBusDaemon *bus_daemon,
     const gchar *object_path,
     GHashTable *immutable_properties,
     GError **error)
+{
+  return _tp_channel_dispatch_operation_new_with_factory (NULL, bus_daemon,
+      object_path, immutable_properties, error);
+}
+
+TpChannelDispatchOperation *
+_tp_channel_dispatch_operation_new_with_factory (TpSimpleClientFactory *factory,
+    TpDBusDaemon *bus_daemon,
+    const gchar *object_path,
+    GHashTable *immutable_properties,
+    GError **error)
 {
   TpChannelDispatchOperation *self;
   gchar *unique_name;
@@ -953,6 +924,7 @@ tp_channel_dispatch_operation_new (TpDBusDaemon *bus_daemon,
         "bus-name", unique_name,
         "object-path", object_path,
         "cdo-properties", immutable_properties,
+        "factory", factory,
         NULL));
 
   g_free (unique_name);
@@ -1222,15 +1194,17 @@ claim_cb (TpChannelDispatchOperation *self,
  * method becomes the handler for the channel.
  *
  * If successful, this method will cause the #TpProxy::invalidated signal
- * to be emitted, in the same wayas for
+ * to be emitted, in the same way as for
  * tp_channel_dispatch_operation_handle_with_async().
  *
  * This method may fail because the dispatch operation has already
- * been completed. Again, see tp_channel_dispatch_operation_claim_async()
+ * been completed. Again, see tp_channel_dispatch_operation_handle_with_async()
  * for more details. The approver MUST NOT attempt to interact with
  * the channels further in this case.
  *
  * Since: 0.11.5
+ * Deprecated: since 0.15.0. Use
+ * tp_channel_dispatch_operation_claim_with_async()
  */
 void
 tp_channel_dispatch_operation_claim_async (
@@ -1260,6 +1234,8 @@ tp_channel_dispatch_operation_claim_async (
  * Returns: %TRUE if the Claim() call was successful, otherwise %FALSE
  *
  * Since: 0.11.5
+ * Deprecated: since 0.15.0. Use
+ * tp_channel_dispatch_operation_claim_with_finish()
  */
 gboolean
 tp_channel_dispatch_operation_claim_finish (
@@ -1283,45 +1259,24 @@ tp_channel_dispatch_operation_claim_finish (
   return TRUE;
 }
 
-/* This can be used by tp-glib to create a new CDO using existing proxies
- * (account, connection, channels) rather than creating new ones. */
-TpChannelDispatchOperation *
-_tp_channel_dispatch_operation_new_with_objects (TpDBusDaemon *bus_daemon,
-    const gchar *object_path,
-    GHashTable *immutable_properties,
-    TpAccount *account,
-    TpConnection *connection,
-    GPtrArray *channels,
-    GError **error)
+/* FIXME: This is temporary solution to share TpChannel objects until
+ * TpSimpleClientFactory can be used for that */
+void
+_tp_channel_dispatch_operation_ensure_channels (TpChannelDispatchOperation *self,
+    GPtrArray *channels)
 {
-  TpChannelDispatchOperation *self;
-  gchar *unique_name;
+  guint i;
 
-  g_return_val_if_fail (bus_daemon != NULL, NULL);
-  g_return_val_if_fail (object_path != NULL, NULL);
-  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+  if (self->priv->channels != NULL || channels == NULL)
+    return;
 
-  if (!tp_dbus_check_valid_object_path (object_path, error))
-    return NULL;
+  /* Do not just ref the GPtrArray because we'll modify its content */
+  self->priv->channels = _tp_g_ptr_array_new_full (channels->len,
+      g_object_unref);
 
-  if (!_tp_dbus_daemon_get_name_owner (bus_daemon, -1,
-      TP_CHANNEL_DISPATCHER_BUS_NAME, &unique_name, error))
-    return NULL;
-
-  self = TP_CHANNEL_DISPATCH_OPERATION (g_object_new (
-        TP_TYPE_CHANNEL_DISPATCH_OPERATION,
-        "dbus-daemon", bus_daemon,
-        "bus-name", unique_name,
-        "object-path", object_path,
-        "cdo-properties", immutable_properties,
-        "account", account,
-        "connection", connection,
-        "channels", channels,
-        NULL));
-
-  g_free (unique_name);
-
-  return self;
+  for (i = 0; i < channels->len; i++)
+    g_ptr_array_add (self->priv->channels,
+        g_object_ref (g_ptr_array_index (channels, i)));
 }
 
 /**
@@ -1407,3 +1362,508 @@ gboolean
 
   return TRUE;
 }
+
+static void
+claim_with_cb (TpChannelDispatchOperation *self,
+    GSimpleAsyncResult *result)
+{
+  TpBaseClient *client;
+
+  client = g_simple_async_result_get_op_res_gpointer (result);
+
+  _tp_base_client_now_handling_channels (client, self->priv->channels);
+
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+}
+
+typedef void (*PrepareCoreAndClaimCb) (TpChannelDispatchOperation *self,
+    GSimpleAsyncResult *result);
+
+typedef struct
+{
+  GSimpleAsyncResult *result;
+  PrepareCoreAndClaimCb callback;
+} PrepareCoreAndClaimCtx;
+
+static PrepareCoreAndClaimCtx *
+prepare_core_and_claim_ctx_new (GSimpleAsyncResult *result,
+    PrepareCoreAndClaimCb callback)
+{
+  PrepareCoreAndClaimCtx *ctx = g_slice_new (PrepareCoreAndClaimCtx);
+
+  ctx->result = g_object_ref (result);
+  ctx->callback = callback;
+  return ctx;
+}
+
+static void
+prepare_core_and_claim_ctx_free (PrepareCoreAndClaimCtx *ctx)
+{
+  g_object_unref (ctx->result);
+  g_slice_free (PrepareCoreAndClaimCtx, ctx);
+}
+
+/* Takes ownership of @error */
+static void
+prepare_core_and_claim_ctx_failed (PrepareCoreAndClaimCtx *ctx,
+    GError *error)
+{
+  g_simple_async_result_take_error (ctx->result, error);
+  g_simple_async_result_complete (ctx->result);
+
+  /* We received a reference on result from the caller and was supposed to
+   * give it back when calling the callback. But as something went wrong, we
+   * terminate the operation ourself and so don't call the callback, so we
+   * have to drop this reference. */
+  g_object_unref (ctx->result);
+  prepare_core_and_claim_ctx_free (ctx);
+}
+
+static void
+prepare_core_claim_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  TpChannelDispatchOperation *self = (TpChannelDispatchOperation *) source;
+  PrepareCoreAndClaimCtx *ctx = user_data;
+  GError *error = NULL;
+
+  if (!tp_channel_dispatch_operation_claim_finish (self, result, &error))
+    {
+      DEBUG ("Failed to Claim %s: %s",
+          tp_proxy_get_object_path (self), error->message);
+
+      prepare_core_and_claim_ctx_failed (ctx, error);
+      return;
+    }
+
+  /* Pass back the ref we got from the caller */
+  ctx->callback (self, ctx->result);
+
+  prepare_core_and_claim_ctx_free (ctx);
+}
+
+static void
+prepare_core_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  TpChannelDispatchOperation *self = (TpChannelDispatchOperation *) source;
+  PrepareCoreAndClaimCtx *ctx = user_data;
+  GError *error = NULL;
+
+  if (!tp_proxy_prepare_finish (self, result, &error))
+    {
+      DEBUG ("Failed to prepare CORE on %s: %s",
+          tp_proxy_get_object_path (self), error->message);
+
+      prepare_core_and_claim_ctx_failed (ctx, error);
+      return;
+    }
+
+  tp_channel_dispatch_operation_claim_async (self, prepare_core_claim_cb, ctx);
+}
+
+/* Prepare CORE feature on @self and then call Claim() on it.
+ * If either the preparation or the call failed, complete @result with the
+ * error.
+ * If everything goes fine call @callback.
+ *
+ * Takes the reference on @result and pass it back to @callback.
+ */
+static void
+prepare_core_and_claim (TpChannelDispatchOperation *self,
+    PrepareCoreAndClaimCb callback,
+    GSimpleAsyncResult *result)
+{
+  GQuark features[] = { TP_CHANNEL_DISPATCH_OPERATION_FEATURE_CORE, 0 };
+  PrepareCoreAndClaimCtx *ctx;
+
+  ctx = prepare_core_and_claim_ctx_new (result, callback);
+
+  /* We have to prepare the CDO to be able to get the list of its channels.
+   * We prepare it *before* calling Claim() as MC will destroy the CDO once it
+   * has been claimed. */
+  tp_proxy_prepare_async (self, features, prepare_core_cb, ctx);
+}
+
+/**
+ * tp_channel_dispatch_operation_claim_with_async:
+ * @self: a #TpChannelDispatchOperation
+ * @client: the #TpBaseClient claiming @self
+ * @callback: a callback to call when the call returns
+ * @user_data: data to pass to @callback
+ *
+ * Called by an approver to claim channels for handling internally.
+ * If this method is called successfully, the process calling this
+ * method becomes the handler for the channel.
+ *
+ * If successful, this method will cause the #TpProxy::invalidated signal
+ * to be emitted, in the same way as for
+ * tp_channel_dispatch_operation_handle_with_async().
+ *
+ * This method may fail because the dispatch operation has already
+ * been completed. Again, see tp_channel_dispatch_operation_handle_with_async()
+ * for more details. The approver MUST NOT attempt to interact with
+ * the channels further in this case.
+ *
+ * This is an improved version of tp_channel_dispatch_operation_claim_async()
+ * as it tells @client about the new channels being handled.
+ *
+ * Since: 0.15.0
+ */
+void
+tp_channel_dispatch_operation_claim_with_async (
+    TpChannelDispatchOperation *self,
+    TpBaseClient *client,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+
+  g_return_if_fail (TP_IS_CHANNEL_DISPATCH_OPERATION (self));
+
+  result = g_simple_async_result_new (G_OBJECT (self),
+      callback, user_data,
+      tp_channel_dispatch_operation_claim_with_async);
+
+  g_simple_async_result_set_op_res_gpointer (result, g_object_ref (client),
+      g_object_unref);
+
+  prepare_core_and_claim (self, claim_with_cb, result);
+}
+
+/**
+ * tp_channel_dispatch_operation_claim_with_finish:
+ * @self: a #TpChannelDispatchOperation
+ * @result: a #GAsyncResult
+ * @error: a #GError to fill
+ *
+ * Finishes an async call to Claim() initiated using
+ * tp_channel_dispatch_operation_claim_with_async().
+ *
+ * Returns: %TRUE if the Claim() call was successful, otherwise %FALSE
+ *
+ * Since: 0.15.0
+ */
+gboolean
+tp_channel_dispatch_operation_claim_with_finish (
+    TpChannelDispatchOperation *self,
+    GAsyncResult *result,
+    GError **error)
+{
+  _tp_implement_finish_void (self, \
+      tp_channel_dispatch_operation_claim_with_async)
+}
+
+static void
+channel_close_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  GError *error = NULL;
+
+  if (!tp_channel_close_finish (TP_CHANNEL (source), result, &error))
+    {
+      DEBUG ("Failed to close %s: %s", tp_proxy_get_object_path (source),
+          error->message);
+
+      g_error_free (error);
+    }
+}
+
+static void
+claim_close_channels_cb (TpChannelDispatchOperation *self,
+    GSimpleAsyncResult *result)
+{
+  guint i;
+
+  for (i = 0; i < self->priv->channels->len; i++)
+    {
+      TpChannel *channel = g_ptr_array_index (self->priv->channels, i);
+
+      tp_channel_close_async (channel, channel_close_cb, NULL);
+    }
+
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+}
+
+/**
+ * tp_channel_dispatch_operation_close_channels_async:
+ * @self: a #TpChannelDispatchOperation
+ * @callback: a callback to call when the request has been satisfied
+ * @user_data: data to pass to @callback
+ *
+ * Called by an approver to claim channels and close them all right away.
+ * If this method is called successfully, @self has been claimed and
+ * tp_channel_close_async() has been called on all of its channels.
+ *
+ * If successful, this method will cause the #TpProxy::invalidated signal
+ * to be emitted, in the same way as for
+ * tp_channel_dispatch_operation_handle_with_async().
+ *
+ * This method may fail because the dispatch operation has already
+ * been completed. Again, see tp_channel_dispatch_operation_handle_with_async()
+ * for more details.
+ *
+ * Since: 0.15.1
+ */
+void
+tp_channel_dispatch_operation_close_channels_async (
+    TpChannelDispatchOperation *self,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+
+  result = g_simple_async_result_new (G_OBJECT (self), callback, user_data,
+      tp_channel_dispatch_operation_close_channels_async);
+
+  prepare_core_and_claim (self, claim_close_channels_cb, result);
+}
+
+/**
+ * tp_channel_dispatch_operation_close_channels_finish:
+ * @self: a #TpChannelDispatchOperation
+ * @result: a #GAsyncResult
+ * @error: a #GError to fill
+ *
+ * Finishes an async operation initiated using
+ * tp_channel_dispatch_operation_close_channels_async().
+ *
+ * Returns: %TRUE if the Claim() call was successful and
+ * Close() has at least been attempted on all the channels, otherwise %FALSE
+ *
+ * Since: 0.15.1
+ */
+gboolean
+tp_channel_dispatch_operation_close_channels_finish (
+    TpChannelDispatchOperation *self,
+    GAsyncResult *result,
+    GError **error)
+{
+  _tp_implement_finish_void (self, \
+      tp_channel_dispatch_operation_close_channels_async)
+}
+
+typedef struct
+{
+  TpChannelGroupChangeReason reason;
+  gchar *message;
+} LeaveChannelsCtx;
+
+static LeaveChannelsCtx *
+leave_channels_ctx_new (TpChannelGroupChangeReason reason,
+    const gchar *message)
+{
+  LeaveChannelsCtx *ctx = g_slice_new (LeaveChannelsCtx);
+
+  ctx->reason = reason;
+  ctx->message = g_strdup (message);
+  return ctx;
+}
+
+static void
+leave_channels_ctx_free (LeaveChannelsCtx *ctx)
+{
+  g_free (ctx->message);
+  g_slice_free (LeaveChannelsCtx, ctx);
+}
+
+static void
+channel_leave_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  GError *error = NULL;
+
+  if (!tp_channel_leave_finish (TP_CHANNEL (source), result, &error))
+    {
+      DEBUG ("Failed to leave %s: %s", tp_proxy_get_object_path (source),
+          error->message);
+
+      g_error_free (error);
+    }
+}
+
+static void
+claim_leave_channels_cb (TpChannelDispatchOperation *self,
+    GSimpleAsyncResult *result)
+{
+  guint i;
+  LeaveChannelsCtx *ctx;
+
+  ctx = g_simple_async_result_get_op_res_gpointer (result);
+
+  for (i = 0; i < self->priv->channels->len; i++)
+    {
+      TpChannel *channel = g_ptr_array_index (self->priv->channels, i);
+
+      tp_channel_leave_async (channel, ctx->reason, ctx->message,
+          channel_leave_cb, NULL);
+    }
+
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+}
+
+/**
+ * tp_channel_dispatch_operation_leave_channels_async:
+ * @self: a #TpChannelDispatchOperation
+ * @reason: the leave reason
+ * @message: the leave message
+ * @callback: a callback to call when the request has been satisfied
+ * @user_data: data to pass to @callback
+ *
+ * Called by an approver to claim channels and leave them all right away.
+ * If this method is called successfully, @self has been claimed and
+ * tp_channel_leave_async() has been called on all of its channels.
+ *
+ * If successful, this method will cause the #TpProxy::invalidated signal
+ * to be emitted, in the same way as for
+ * tp_channel_dispatch_operation_handle_with_async().
+ *
+ * This method may fail because the dispatch operation has already
+ * been completed. Again, see tp_channel_dispatch_operation_handle_with_async()
+ * for more details.
+ *
+ * Since: 0.15.2
+ */
+void
+tp_channel_dispatch_operation_leave_channels_async (
+    TpChannelDispatchOperation *self,
+    TpChannelGroupChangeReason reason,
+    const gchar *message,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+
+  result = g_simple_async_result_new (G_OBJECT (self), callback, user_data,
+      tp_channel_dispatch_operation_leave_channels_async);
+
+  g_simple_async_result_set_op_res_gpointer (result,
+      leave_channels_ctx_new (reason, message),
+      (GDestroyNotify) leave_channels_ctx_free);
+
+  prepare_core_and_claim (self, claim_leave_channels_cb, result);
+}
+
+/**
+ * tp_channel_dispatch_operation_leave_channels_finish:
+ * @self: a #TpChannelDispatchOperation
+ * @result: a #GAsyncResult
+ * @error: a #GError to fill
+ *
+ * Finishes an async operation initiated using
+ * tp_channel_dispatch_operation_leave_channels_async().
+ *
+ * Returns: %TRUE if the Claim() call was successful and
+ * tp_channel_leave_async() has at least been attempted on all the
+ * channels, otherwise %FALSE
+ *
+ * Since: 0.15.2
+ */
+gboolean
+tp_channel_dispatch_operation_leave_channels_finish (
+    TpChannelDispatchOperation *self,
+    GAsyncResult *result,
+    GError **error)
+{
+  _tp_implement_finish_void (self, \
+      tp_channel_dispatch_operation_leave_channels_async)
+}
+
+static void
+channel_destroy_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  GError *error = NULL;
+
+  if (!tp_channel_destroy_finish (TP_CHANNEL (source), result, &error))
+    {
+      DEBUG ("Failed to destroy %s: %s", tp_proxy_get_object_path (source),
+          error->message);
+
+      g_error_free (error);
+    }
+}
+
+static void
+claim_destroy_channels_cb (TpChannelDispatchOperation *self,
+    GSimpleAsyncResult *result)
+{
+  guint i;
+
+  for (i = 0; i < self->priv->channels->len; i++)
+    {
+      TpChannel *channel = g_ptr_array_index (self->priv->channels, i);
+
+      tp_channel_destroy_async (channel, channel_destroy_cb, NULL);
+    }
+
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+}
+
+/**
+ * tp_channel_dispatch_operation_destroy_channels_async:
+ * @self: a #TpChannelDispatchOperation
+ * @callback: a callback to call when the request has been satisfied
+ * @user_data: data to pass to @callback
+ *
+ * Called by an approver to claim channels and destroy them all right away.
+ * If this method is called successfully, @self has been claimed and
+ * tp_channel_destroy_async() has been called on all of its channels.
+ *
+ * If successful, this method will cause the #TpProxy::invalidated signal
+ * to be emitted, in the same way as for
+ * tp_channel_dispatch_operation_handle_with_async().
+ *
+ * This method may fail because the dispatch operation has already
+ * been completed. Again, see tp_channel_dispatch_operation_handle_with_async()
+ * for more details.
+ *
+ * Since: 0.15.2
+ */
+void
+tp_channel_dispatch_operation_destroy_channels_async (
+    TpChannelDispatchOperation *self,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+
+  result = g_simple_async_result_new (G_OBJECT (self), callback, user_data,
+      tp_channel_dispatch_operation_destroy_channels_async);
+
+  prepare_core_and_claim (self, claim_destroy_channels_cb, result);
+}
+
+/**
+ * tp_channel_dispatch_operation_destroy_channels_finish:
+ * @self: a #TpChannelDispatchOperation
+ * @result: a #GAsyncResult
+ * @error: a #GError to fill
+ *
+ * Finishes an async operation initiated using
+ * tp_channel_dispatch_operation_destroy_channels_async().
+ *
+ * Returns: %TRUE if the Claim() call was successful and
+ * tp_channel_destroy_async() has at least been attempted on all the
+ * channels, otherwise %FALSE
+ *
+ * Since: 0.15.2
+ */
+gboolean
+tp_channel_dispatch_operation_destroy_channels_finish (
+    TpChannelDispatchOperation *self,
+    GAsyncResult *result,
+    GError **error)
+{
+  _tp_implement_finish_void (self, \
+      tp_channel_dispatch_operation_destroy_channels_async)
+}
