@@ -21,6 +21,7 @@
 
 #include "telepathy-glib/channel-request.h"
 
+#include <telepathy-glib/account-manager.h>
 #include <telepathy-glib/automatic-proxy-factory.h>
 #include <telepathy-glib/channel.h>
 #include <telepathy-glib/connection.h>
@@ -34,7 +35,9 @@
 #define DEBUG_FLAG TP_DEBUG_DISPATCHER
 #include "telepathy-glib/dbus-internal.h"
 #include "telepathy-glib/debug-internal.h"
-
+#include "telepathy-glib/deprecated-internal.h"
+#include "telepathy-glib/proxy-internal.h"
+#include "telepathy-glib/simple-client-factory-internal.h"
 #include "telepathy-glib/_gen/tp-cli-channel-request-body.h"
 
 #include "_gen/signals-marshal.h"
@@ -102,6 +105,9 @@ enum {
 enum {
   PROP_CHANNEL_FACTORY = 1,
   PROP_IMMUTABLE_PROPERTIES,
+  PROP_ACCOUNT,
+  PROP_USER_ACTION_TIME,
+  PROP_PREFERRED_HANDLER,
   PROP_HINTS,
 };
 
@@ -112,6 +118,7 @@ struct _TpChannelRequestPrivate {
 
     TpClientChannelFactory *channel_factory;
     gboolean succeeded_with_chan_fired;
+    TpAccount *account;
 };
 
 G_DEFINE_TYPE (TpChannelRequest, tp_channel_request, TP_TYPE_PROXY)
@@ -167,6 +174,20 @@ tp_channel_request_get_property (GObject *object,
         g_value_set_boxed (value, self->priv->immutable_properties);
         break;
 
+      case PROP_ACCOUNT:
+        g_value_set_object (value, tp_channel_request_get_account (self));
+        break;
+
+      case PROP_USER_ACTION_TIME:
+        g_value_set_int64 (value,
+            tp_channel_request_get_user_action_time (self));
+        break;
+
+      case PROP_PREFERRED_HANDLER:
+        g_value_set_string (value,
+            tp_channel_request_get_preferred_handler (self));
+        break;
+
       case PROP_HINTS:
         g_value_set_boxed (value, tp_channel_request_get_hints (self));
         break;
@@ -224,14 +245,12 @@ tp_channel_request_succeeded_with_channel_cb (TpChannelRequest *self,
     gpointer unused G_GNUC_UNUSED,
     GObject *object G_GNUC_UNUSED)
 {
-  TpDBusDaemon *dbus;
   TpConnection *connection;
   TpChannel *channel;
   GError *error = NULL;
 
-  dbus = tp_proxy_get_dbus_daemon (self);
-
-  connection = tp_connection_new (dbus, NULL, conn_path, &error);
+  connection = tp_simple_client_factory_ensure_connection (
+      tp_proxy_get_factory (self), conn_path, NULL, &error);
   if (connection == NULL)
     {
       DEBUG ("Failed to create TpConnection: %s", error->message);
@@ -239,13 +258,17 @@ tp_channel_request_succeeded_with_channel_cb (TpChannelRequest *self,
       return;
     }
 
-  channel = tp_client_channel_factory_create_channel (
-      self->priv->channel_factory, connection, chan_path, chan_props, &error);
+  if (self->priv->channel_factory != NULL)
+    channel = tp_client_channel_factory_create_channel (
+        self->priv->channel_factory, connection, chan_path, chan_props, &error);
+  else
+    channel = tp_simple_client_factory_ensure_channel (tp_proxy_get_factory (self),
+        connection, chan_path, chan_props, &error);
   if (channel == NULL)
     {
       DEBUG ("Failed to create TpChannel: %s", error->message);
       g_error_free (error);
-      g_object_unref (channel);
+      g_object_unref (connection);
       return;
     }
 
@@ -272,11 +295,7 @@ tp_channel_request_constructed (GObject *object)
 
   g_return_if_fail (tp_proxy_get_dbus_daemon (self) != NULL);
 
-  if (self->priv->channel_factory == NULL)
-    {
-      self->priv->channel_factory = TP_CLIENT_CHANNEL_FACTORY (
-          tp_automatic_proxy_factory_dup ());
-    }
+  _tp_proxy_ensure_factory (self, NULL);
 
   sc = tp_cli_channel_request_connect_to_failed (self,
       tp_channel_request_failed_cb, NULL, NULL, NULL, &error);
@@ -321,6 +340,7 @@ tp_channel_request_dispose (GObject *object)
   tp_clear_pointer (&self->priv->immutable_properties, g_hash_table_unref);
 
   tp_clear_object (&self->priv->channel_factory);
+  tp_clear_object (&self->priv->account);
 
   if (dispose != NULL)
     dispose (object);
@@ -356,6 +376,7 @@ tp_channel_request_class_init (TpChannelRequestClass *klass)
    * If no channel factory is specified then #TpAutomaticProxyFactory is used.
    *
    * Since: 0.13.14
+   * Deprecated: since 0.15.5. Use #TpProxy:factory instead.
    */
   param_spec = g_param_spec_object ("channel-factory", "Channel factory",
       "Object implementing TpClientChannelFactoryInterface",
@@ -386,6 +407,57 @@ tp_channel_request_class_init (TpChannelRequestClass *klass)
   g_object_class_install_property (object_class, PROP_IMMUTABLE_PROPERTIES,
       param_spec);
 
+  /**
+   * TpChannelRequest:account:
+   *
+   * The #TpAccount on which this request was made, not guaranteed
+   * to be prepared.
+   *
+   * Read-only.
+   *
+   * Since: 0.15.3
+   */
+  param_spec = g_param_spec_object ("account", "Account", "Account",
+      TP_TYPE_ACCOUNT,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_ACCOUNT, param_spec);
+
+  /**
+   * TpChannelRequest:user-action-time:
+   *
+   * The time at which user action occurred, or
+   * #TP_USER_ACTION_TIME_NOT_USER_ACTION if this channel request is
+   * for some reason not involving user action.
+   *
+   * Read-only.
+   *
+   * Since: 0.15.3
+   */
+  param_spec = g_param_spec_int64 ("user-action-time", "UserActionTime",
+      "UserActionTime",
+      0, G_MAXINT64, 0,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_USER_ACTION_TIME,
+      param_spec);
+
+  /**
+   * TpChannelRequest:preferred-handler:
+   *
+   * Either the well-known bus name (starting with #TP_CLIENT_BUS_NAME_BASE)
+   * of the preferred handler for this channel request,
+   * or %NULL to indicate that any handler would be acceptable.
+   *
+   * Read-only.
+   *
+   * Since: 0.15.3
+   */
+  param_spec = g_param_spec_string ("preferred-handler", "PreferredHandler",
+      "PreferredHandler",
+      NULL,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_PREFERRED_HANDLER,
+      param_spec);
+
   /**
    * TpChannelRequest:hints:
    *
@@ -432,10 +504,10 @@ tp_channel_request_class_init (TpChannelRequestClass *klass)
    * @channel will be %NULL. When using newer versions, they will be correctly
    * set to the newly-created channel, and the connection which owns it.
    *
-   * The #TpChannel is created using #TpChannelRequest:channel-factory but
-   * the features of the factory are NOT prepared. It's up to the user to
-   * prepare the features returned by
-   * tp_client_channel_factory_dup_channel_features() himself.
+   * The #TpChannel is created using #TpChannelRequest:channel-factory or
+   * #TpProxy:factory but the features of the factory are NOT prepared.
+   * It's up to the user to prepare the features returned by
+   * tp_simple_client_factory_dup_channel_features() himself.
    *
    * Since: 0.13.14
    */
@@ -503,6 +575,17 @@ tp_channel_request_new (TpDBusDaemon *bus_daemon,
     const gchar *object_path,
     GHashTable *immutable_properties,
     GError **error)
+{
+  return _tp_channel_request_new_with_factory (NULL, bus_daemon, object_path,
+      immutable_properties, error);
+}
+
+TpChannelRequest *
+_tp_channel_request_new_with_factory (TpSimpleClientFactory *factory,
+    TpDBusDaemon *bus_daemon,
+    const gchar *object_path,
+    GHashTable *immutable_properties,
+    GError **error)
 {
   TpChannelRequest *self;
   gchar *unique_name;
@@ -524,6 +607,7 @@ tp_channel_request_new (TpDBusDaemon *bus_daemon,
         "bus-name", unique_name,
         "object-path", object_path,
         "immutable-properties", immutable_properties,
+        "factory", factory,
         NULL));
 
   g_free (unique_name);
@@ -540,13 +624,23 @@ tp_channel_request_new (TpDBusDaemon *bus_daemon,
  * Change the value of the #TpChannelRequest:channel-factory property.
  *
  * Since: 0.13.14
+ * Deprecated: since 0.15.5. Use #TpProxy:factory instead.
  */
 void
 tp_channel_request_set_channel_factory (TpChannelRequest *self,
     TpClientChannelFactory *factory)
+{
+  _tp_channel_request_set_channel_factory (self, factory);
+}
+
+void
+_tp_channel_request_set_channel_factory (TpChannelRequest *self,
+    TpClientChannelFactory *factory)
 {
   tp_clear_object (&self->priv->channel_factory);
-  self->priv->channel_factory = g_object_ref (factory);
+
+  if (factory != NULL)
+    self->priv->channel_factory = g_object_ref (factory);
   g_object_notify (G_OBJECT (self), "channel-factory");
 }
 
@@ -569,6 +663,96 @@ tp_channel_request_get_immutable_properties (TpChannelRequest *self)
   return self->priv->immutable_properties;
 }
 
+void
+_tp_channel_request_ensure_immutable_properties (TpChannelRequest *self,
+    GHashTable *immutable_properties)
+{
+  if (self->priv->immutable_properties == NULL && immutable_properties != NULL)
+    {
+      self->priv->immutable_properties = g_hash_table_ref (immutable_properties);
+      g_object_notify ((GObject *) self, "immutable-properties");
+    }
+}
+
+/**
+ * tp_channel_request_get_account:
+ * @self: a #tpchannelrequest
+ *
+ * Return the value of the #TpChannelRequest:account construct-only property
+ *
+ * returns: (transfer none): the value of #TpChannelRequest:account
+ *
+ * since: 0.15.3
+ */
+TpAccount *
+tp_channel_request_get_account (TpChannelRequest *self)
+{
+  g_return_val_if_fail (TP_IS_CHANNEL_REQUEST (self), NULL);
+
+  /* lazily initialize self->priv->account */
+  if (self->priv->account == NULL)
+    {
+      const gchar *path;
+
+      if (self->priv->immutable_properties == NULL)
+        return NULL;
+
+      path = tp_asv_get_object_path (self->priv->immutable_properties,
+          TP_PROP_CHANNEL_REQUEST_ACCOUNT);
+      if (path == NULL)
+        return NULL;
+
+      self->priv->account = tp_simple_client_factory_ensure_account (
+          tp_proxy_get_factory (self), path, NULL, NULL);
+    }
+
+  return self->priv->account;
+}
+
+/**
+ * tp_channel_request_get_user_action_time:
+ * @self: a #tpchannelrequest
+ *
+ * return the #TpChannelRequest:user-action-time construct-only property
+ *
+ * returns: the value of #TpChannelRequest:user-action-time
+ *
+ * since: 0.15.3
+ */
+gint64
+tp_channel_request_get_user_action_time (TpChannelRequest *self)
+{
+  g_return_val_if_fail (TP_IS_CHANNEL_REQUEST (self), 0);
+
+  if (self->priv->immutable_properties == NULL)
+    return 0;
+
+  return tp_asv_get_int64 (self->priv->immutable_properties,
+      TP_PROP_CHANNEL_REQUEST_USER_ACTION_TIME, NULL);
+}
+
+/**
+ * tp_channel_request_get_preferred_handler:
+ * @self: a #tpchannelrequest
+ *
+ * return the #TpChannelRequest:preferred-handler construct-only property
+ *
+ * returns: the value of #TpChannelRequest:preferred-handler
+ *
+ * since: 0.15.3
+ */
+const gchar *
+tp_channel_request_get_preferred_handler (TpChannelRequest *self)
+{
+  g_return_val_if_fail (TP_IS_CHANNEL_REQUEST (self), NULL);
+
+  if (self->priv->immutable_properties == NULL)
+    return NULL;
+
+  return tp_asv_get_string (self->priv->immutable_properties,
+      TP_PROP_CHANNEL_REQUEST_PREFERRED_HANDLER);
+}
+
 /**
  * tp_channel_request_get_hints:
  * @self: a #TpChannelRequest
