@@ -62,6 +62,7 @@
 
 #define DEBUG_FLAG TP_DEBUG_CHANNEL
 #include "telepathy-glib/debug-internal.h"
+#include "telepathy-glib/automatic-client-factory-internal.h"
 
 #include "_gen/signals-marshal.h"
 
@@ -80,6 +81,9 @@ struct _TpTextChannelPrivate
   /* queue of owned TpSignalledMessage */
   GQueue *pending_messages;
   gboolean got_initial_messages;
+
+  gboolean is_sms_channel;
+  gboolean sms_flash;
 };
 
 enum
@@ -88,6 +92,8 @@ enum
   PROP_MESSAGE_PART_SUPPORT_FLAGS,
   PROP_DELIVERY_REPORTING_SUPPORT,
   PROP_MESSAGE_TYPES,
+  PROP_IS_SMS_CHANNEL,
+  PROP_SMS_FLASH,
 };
 
 enum /* signals */
@@ -144,6 +150,14 @@ tp_text_channel_get_property (GObject *object,
             tp_text_channel_get_message_types (self));
         break;
 
+      case PROP_IS_SMS_CHANNEL:
+        g_value_set_boolean (value, tp_text_channel_is_sms_channel (self));
+        break;
+
+      case PROP_SMS_FLASH:
+        g_value_set_boolean (value, tp_text_channel_get_sms_flash (self));
+        break;
+
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
         break;
@@ -335,6 +349,10 @@ tp_text_channel_constructed (GObject *obj)
           tp_proxy_get_object_path (self), err->message);
       g_error_free (err);
     }
+
+  /* SMS */
+  self->priv->sms_flash = tp_asv_get_boolean (props,
+      TP_PROP_CHANNEL_INTERFACE_SMS_FLASH, NULL);
 }
 
 static void
@@ -580,6 +598,17 @@ got_pending_senders_contact (TpTextChannel *self,
     }
 }
 
+static void
+free_parts_list (GList *parts_list)
+{
+  GList *l;
+
+  for (l = parts_list; l != NULL; l = g_list_next (l))
+    g_boxed_free (TP_ARRAY_TYPE_MESSAGE_PART_LIST, l->data);
+
+  g_list_free (parts_list);
+}
+
 static void
 got_pending_senders_contact_by_handle_cb (TpConnection *connection,
     guint n_contacts,
@@ -590,12 +619,15 @@ got_pending_senders_contact_by_handle_cb (TpConnection *connection,
     gpointer user_data,
     GObject *weak_object)
 {
-  TpTextChannel *self = (TpTextChannel *) weak_object;
+  GSimpleAsyncResult *result = (GSimpleAsyncResult *) weak_object;
   GList *parts_list = user_data;
+  TpTextChannel *self = TP_TEXT_CHANNEL (g_async_result_get_source_object (
+        G_ASYNC_RESULT (result)));
 
   if (error != NULL)
     {
       DEBUG ("Failed to prepare TpContact: %s", error->message);
+      g_simple_async_result_set_from_error (result, error);
       goto out;
     }
 
@@ -607,8 +639,9 @@ got_pending_senders_contact_by_handle_cb (TpConnection *connection,
   got_pending_senders_contact (self, parts_list, n_contacts, contacts);
 
 out:
-  _tp_proxy_set_feature_prepared (TP_PROXY (self),
-      TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, TRUE);
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+  g_object_unref (self);
 }
 
 static void
@@ -621,12 +654,15 @@ got_pending_senders_contact_by_id_cb (TpConnection *connection,
     gpointer user_data,
     GObject *weak_object)
 {
-  TpTextChannel *self = (TpTextChannel *) weak_object;
+  GSimpleAsyncResult *result = (GSimpleAsyncResult *) weak_object;
   GList *parts_list = user_data;
+  TpTextChannel *self = TP_TEXT_CHANNEL (g_async_result_get_source_object (
+        G_ASYNC_RESULT (result)));
 
   if (error != NULL)
     {
       DEBUG ("Failed to prepare TpContact: %s", error->message);
+      g_simple_async_result_set_from_error (result, error);
       goto out;
     }
 
@@ -646,20 +682,9 @@ got_pending_senders_contact_by_id_cb (TpConnection *connection,
   got_pending_senders_contact (self, parts_list, n_contacts, contacts);
 
 out:
-  _tp_proxy_set_feature_prepared (TP_PROXY (self),
-      TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, TRUE);
-}
-
-static void
-free_parts_list (gpointer data)
-{
-  GList *parts_list = data;
-  GList *l;
-
-  for (l = parts_list; l != NULL; l = g_list_next (l))
-    g_boxed_free (TP_ARRAY_TYPE_MESSAGE_PART_LIST, l->data);
-
-  g_list_free (parts_list);
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+  g_object_unref (self);
 }
 
 /* There is no TP_ARRAY_TYPE_PENDING_TEXT_MESSAGE_LIST_LIST (fdo #32433) */
@@ -673,11 +698,13 @@ get_pending_messages_cb (TpProxy *proxy,
     gpointer user_data,
     GObject *weak_object)
 {
-  TpTextChannel *self = user_data;
+  TpTextChannel *self = (TpTextChannel *) proxy;
+  GSimpleAsyncResult *result = user_data;
   guint i;
   GPtrArray *messages;
   TpIntSet *senders;
-  GList *parts_list = NULL;
+  /* Messages we're waiting for a sender for */
+  GQueue outstanding = G_QUEUE_INIT;
   GPtrArray *sender_ids;
 
   self->priv->got_initial_messages = TRUE;
@@ -686,8 +713,10 @@ get_pending_messages_cb (TpProxy *proxy,
     {
       DEBUG ("Failed to get PendingMessages property: %s", error->message);
 
-      _tp_proxy_set_feature_prepared (proxy,
-          TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, FALSE);
+      g_simple_async_result_set_error (result, error->domain, error->code,
+          "Failed to get PendingMessages property: %s", error->message);
+
+      g_simple_async_result_complete (result);
       return;
     }
 
@@ -695,8 +724,10 @@ get_pending_messages_cb (TpProxy *proxy,
     {
       DEBUG ("PendingMessages property is of the wrong type");
 
-      _tp_proxy_set_feature_prepared (proxy,
-          TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, FALSE);
+      g_simple_async_result_set_error (result, TP_ERRORS, TP_ERROR_CONFUSED,
+          "PendingMessages property is of the wrong type");
+
+      g_simple_async_result_complete (result);
       return;
     }
 
@@ -733,33 +764,33 @@ get_pending_messages_cb (TpProxy *proxy,
       if (sender_id != NULL)
         g_ptr_array_add (sender_ids, (gpointer) sender_id);
 
-      /* We'll revert the list below when requesting the TpContact objects */
-      parts_list = g_list_prepend (parts_list, copy_parts (parts));
+      g_queue_push_tail (&outstanding, copy_parts (parts));
     }
 
   if (tp_intset_size (senders) == 0)
     {
-      _tp_proxy_set_feature_prepared (proxy,
-          TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, TRUE);
+      g_simple_async_result_complete (result);
     }
   else
     {
-      TpConnection *conn;
-
-      parts_list = g_list_reverse (parts_list);
-
-      conn = tp_channel_borrow_connection (TP_CHANNEL (proxy));
+      TpConnection *conn = tp_channel_borrow_connection (TP_CHANNEL (proxy));
 
       DEBUG ("Pending messages may be re-ordered, please fix CM (%s)",
           tp_proxy_get_object_path (conn));
 
-      if (sender_ids->len == g_list_length (parts_list))
+      /* If we have an identifier for the sender of every outstanding message,
+       * use those rather than handles to get the contacts. (There may be
+       * duplicates, but telepathy-glib copes.)
+       *
+       * Ownership of the parts list in 'outstanding' is transferred to the
+       * callbacks.
+       */
+      if (sender_ids->len == outstanding.length)
         {
-          /* Use the sender ID rather than the handles */
           tp_connection_get_contacts_by_id (conn, sender_ids->len,
               (const gchar * const *) sender_ids->pdata,
-              0, NULL, got_pending_senders_contact_by_id_cb, parts_list,
-              free_parts_list, G_OBJECT (self));
+              0, NULL, got_pending_senders_contact_by_id_cb, outstanding.head,
+              (GDestroyNotify) free_parts_list, g_object_ref (result));
         }
       else
         {
@@ -767,11 +798,12 @@ get_pending_messages_cb (TpProxy *proxy,
 
           tp_connection_get_contacts_by_handle (conn, tmp->len,
               (TpHandle *) tmp->data,
-              0, NULL, got_pending_senders_contact_by_handle_cb, parts_list,
-              free_parts_list, G_OBJECT (self));
+              0, NULL, got_pending_senders_contact_by_handle_cb, outstanding.head,
+              (GDestroyNotify) free_parts_list, g_object_ref (result));
 
           g_array_unref (tmp);
         }
+
     }
 
   tp_intset_destroy (senders);
@@ -779,10 +811,17 @@ get_pending_messages_cb (TpProxy *proxy,
 }
 
 static void
-tp_text_channel_prepare_pending_messages (TpProxy *proxy)
+tp_text_channel_prepare_pending_messages_async (TpProxy *proxy,
+    const TpProxyFeature *feature,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
 {
   TpChannel *channel = (TpChannel *) proxy;
   GError *error = NULL;
+  GSimpleAsyncResult *result;
+
+  result = g_simple_async_result_new ((GObject *) proxy, callback, user_data,
+      tp_text_channel_prepare_pending_messages_async);
 
   tp_cli_channel_interface_messages_connect_to_message_received (channel,
       message_received_cb, proxy, NULL, G_OBJECT (proxy), &error);
@@ -804,19 +843,102 @@ tp_text_channel_prepare_pending_messages (TpProxy *proxy)
 
   tp_cli_dbus_properties_call_get (proxy, -1,
       TP_IFACE_CHANNEL_INTERFACE_MESSAGES, "PendingMessages",
-      get_pending_messages_cb, proxy, NULL, G_OBJECT (proxy));
+      get_pending_messages_cb, result, g_object_unref, G_OBJECT (proxy));
 
   return;
 
 fail:
-  g_error_free (error);
+  g_simple_async_result_take_error (result, error);
 
-  _tp_proxy_set_feature_prepared (proxy,
-      TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, FALSE);
+  g_simple_async_result_complete_in_idle (result);
+  g_object_unref (result);
+}
+
+static void
+get_sms_channel_cb (TpProxy *proxy,
+    const GValue *value,
+    const GError *error,
+    gpointer user_data,
+    GObject *weak_object)
+{
+  TpTextChannel *self = (TpTextChannel *) proxy;
+  GSimpleAsyncResult *result = user_data;
+
+  if (error != NULL)
+    {
+      DEBUG ("Failed to get SMSChannel property: %s", error->message);
+
+      g_simple_async_result_set_error (result, error->domain, error->code,
+          "Failed to get SMSChannel property: %s", error->message);
+      goto out;
+    }
+
+  if (!G_VALUE_HOLDS (value, G_TYPE_BOOLEAN))
+    {
+      DEBUG ("SMSChannel property is of the wrong type");
+
+      g_simple_async_result_set_error (result, TP_ERRORS, TP_ERROR_CONFUSED,
+          "SMSChannel property is of the wrong type");
+      goto out;
+    }
+
+  self->priv->is_sms_channel = g_value_get_boolean (value);
+
+  /* self->priv->is_sms_channel is set to FALSE by default, so only notify the
+   * property change is it is now set to TRUE. */
+  if (self->priv->is_sms_channel)
+    g_object_notify (G_OBJECT (self), "is-sms-channel");
+
+out:
+  g_simple_async_result_complete (result);
+  g_object_unref (result);
+}
+
+static void
+sms_channel_changed_cb (TpChannel *proxy,
+    gboolean sms,
+    gpointer user_data,
+    GObject *weak_object)
+{
+  TpTextChannel *self = (TpTextChannel *) proxy;
+
+  if (self->priv->is_sms_channel == sms)
+    return;
+
+  self->priv->is_sms_channel = sms;
+  g_object_notify (weak_object, "is-sms-channel");
+}
+
+static void
+tp_text_channel_prepare_sms_async (TpProxy *proxy,
+    const TpProxyFeature *feature,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+  GError *error = NULL;
+
+  result = g_simple_async_result_new ((GObject *) proxy, callback, user_data,
+      tp_text_channel_prepare_sms_async);
+
+  tp_cli_channel_interface_sms_connect_to_sms_channel_changed (
+      (TpChannel *) proxy, sms_channel_changed_cb, NULL, NULL,
+      G_OBJECT (proxy), &error);
+  if (error != NULL)
+    {
+      WARNING ("Failed to connect to SMS.SMSChannelChanged: %s",
+          error->message);
+      g_error_free (error);
+    }
+
+  tp_cli_dbus_properties_call_get (proxy, -1,
+      TP_IFACE_CHANNEL_INTERFACE_SMS, "SMSChannel",
+      get_sms_channel_cb, result, NULL, G_OBJECT (proxy));
 }
 
 enum {
     FEAT_PENDING_MESSAGES,
+    FEAT_SMS,
     N_FEAT
 };
 
@@ -824,14 +946,22 @@ static const TpProxyFeature *
 tp_text_channel_list_features (TpProxyClass *cls G_GNUC_UNUSED)
 {
   static TpProxyFeature features[N_FEAT + 1] = { { 0 } };
+  static GQuark need_sms[2] = {0, 0};
 
   if (G_LIKELY (features[0].name != 0))
     return features;
 
   features[FEAT_PENDING_MESSAGES].name =
     TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES;
-  features[FEAT_PENDING_MESSAGES].start_preparing =
-    tp_text_channel_prepare_pending_messages;
+  features[FEAT_PENDING_MESSAGES].prepare_async =
+    tp_text_channel_prepare_pending_messages_async;
+
+  features[FEAT_SMS].name =
+    TP_TEXT_CHANNEL_FEATURE_SMS;
+  features[FEAT_SMS].prepare_async =
+    tp_text_channel_prepare_sms_async;
+  need_sms[0] = TP_IFACE_QUARK_CHANNEL_INTERFACE_SMS;
+  features[FEAT_SMS].interfaces_needed = need_sms;
 
   /* assert that the terminator at the end is there */
   g_assert (features[N_FEAT].name == 0);
@@ -916,6 +1046,42 @@ tp_text_channel_class_init (TpTextChannelClass *klass)
   g_object_class_install_property (gobject_class,
       PROP_MESSAGE_TYPES, param_spec);
 
+ /**
+   * TpTextChannel:is-sms-channel:
+   *
+   * %TRUE if messages sent and received on this channel are transmitted
+   * via SMS.
+   *
+   * This property is not guaranteed to have a meaningful value until
+   * TP_TEXT_CHANNEL_FEATURE_SMS has been prepared.
+   *
+   * Since: 0.15.1
+   */
+  param_spec = g_param_spec_boolean ("is-sms-channel",
+      "is SMS channel",
+      "The SMS.SMSChannel property of the channel",
+      FALSE,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (gobject_class, PROP_IS_SMS_CHANNEL,
+      param_spec);
+
+ /**
+   * TpTextChannel:sms-flash:
+   *
+   * %TRUE if this channel is exclusively for receiving class 0 SMSes
+   * (and no SMSes can be sent using tp_text_channel_send_message_async()
+   * on this channel). If %FALSE, no incoming class 0 SMSes will appear
+   * on this channel.
+   *
+   * Since: 0.15.1
+   */
+  param_spec = g_param_spec_boolean ("sms-flash",
+      "SMS flash",
+      "The SMS.Flash property of the channel",
+      FALSE,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (gobject_class, PROP_SMS_FLASH, param_spec);
+
   /**
    * TpTextChannel::message-received
    * @self: the #TpTextChannel
@@ -924,6 +1090,10 @@ tp_text_channel_class_init (TpTextChannelClass *klass)
    * The ::message-received signal is emitted when a new message has been
    * received on @self.
    *
+   * The same @message object will be used by the
+   * #TpTextChannel::pending-message-removed signal once @message has been
+   * acked so you can simply compare pointers to identify the message.
+   *
    * Note that this signal is only fired once the
    * #TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES has been prepared.
    *
@@ -1017,6 +1187,18 @@ tp_text_channel_new (TpConnection *conn,
     const gchar *object_path,
     const GHashTable *immutable_properties,
     GError **error)
+{
+  return _tp_text_channel_new_with_factory (NULL, conn, object_path,
+      immutable_properties, error);
+}
+
+TpTextChannel *
+_tp_text_channel_new_with_factory (
+    TpSimpleClientFactory *factory,
+    TpConnection *conn,
+    const gchar *object_path,
+    const GHashTable *immutable_properties,
+    GError **error)
 {
   TpProxy *conn_proxy = (TpProxy *) conn;
 
@@ -1034,6 +1216,7 @@ tp_text_channel_new (TpConnection *conn,
        "object-path", object_path,
        "handle-type", (guint) TP_UNKNOWN_HANDLE_TYPE,
        "channel-properties", immutable_properties,
+       "factory", factory,
        NULL);
 }
 
@@ -1253,6 +1436,10 @@ acknowledge_pending_messages_cb (TpChannel *channel,
  * You can then call tp_text_channel_ack_messages_finish() to get the
  * result of the operation.
  *
+ * You should use the #TpSignalledMessage received from
+ * tp_text_channel_get_pending_messages() or the
+ * #TpTextChannel::message-received signal.
+ *
  * See tp_text_channel_ack_message_async() about acknowledging messages.
  *
  * Since: 0.13.10
@@ -1347,6 +1534,10 @@ tp_text_channel_ack_messages_finish (TpTextChannel *self,
  * pending-message queue and so the #TpTextChannel::pending-message-removed
  * signal is fired.
  *
+ * You should use the #TpSignalledMessage received from
+ * tp_text_channel_get_pending_messages() or the
+ * #TpTextChannel::message-received signal.
+ *
  * Since: 0.13.10
  */
 void
@@ -1522,3 +1713,260 @@ tp_text_channel_supports_message_type (TpTextChannel *self,
 
   return FALSE;
 }
+
+/**
+ * TP_TEXT_CHANNEL_FEATURE_SMS:
+ *
+ * Expands to a call to a function that returns a quark representing the
+ * SMS feature of a #TpTextChannel.
+ *
+ * When this feature is prepared, the TpTextChannel:is-sms-channel property
+ * will have a meaningful value and will be updated when needed.
+ *
+ * One can ask for a feature to be prepared using the
+ * tp_proxy_prepare_async() function, and waiting for it to callback.
+ *
+ * Since: 0.15.1
+ */
+GQuark
+tp_text_channel_get_feature_quark_sms (void)
+{
+  return g_quark_from_static_string ("tp-text-channel-feature-sms");
+}
+
+/**
+ * tp_text_channel_is_sms_channel:
+ * @self: a #TpTextChannel
+ *
+ * Return the #TpTextChannel:is-sms-channel property
+ *
+ * Returns: the value of #TpTextChannel:is-sms-channel property
+ *
+ * Since: 0.15.1
+ */
+gboolean
+tp_text_channel_is_sms_channel (TpTextChannel *self)
+{
+  g_return_val_if_fail (TP_IS_TEXT_CHANNEL (self), FALSE);
+
+  return self->priv->is_sms_channel;
+}
+
+/**
+ * tp_text_channel_get_sms_flash:
+ * @self: a #TpTextChannel
+ *
+ * Return the #TpTextChannel:sms-flash property
+ *
+ * Returns: the value of #TpTextChannel:sms-flash property
+ *
+ * Since: 0.15.1
+ */
+gboolean
+tp_text_channel_get_sms_flash (TpTextChannel *self)
+{
+  g_return_val_if_fail (TP_IS_TEXT_CHANNEL (self), FALSE);
+
+  return self->priv->sms_flash;
+}
+
+typedef struct
+{
+  guint chunks_required;
+  gint remaining_characters;
+  gint estimated_cost;
+} GetSmsLengthReturn;
+
+static GetSmsLengthReturn *
+get_sms_length_return_new (guint chunks_required,
+    gint remaining_characters,
+    gint estimated_cost)
+{
+  GetSmsLengthReturn *result = g_slice_new (GetSmsLengthReturn);
+
+  result->chunks_required = chunks_required;
+  result->remaining_characters = remaining_characters;
+  result->estimated_cost = estimated_cost;
+
+  return result;
+}
+
+static void
+get_sms_length_return_free (GetSmsLengthReturn *r)
+{
+  g_slice_free (GetSmsLengthReturn, r);
+}
+
+static void
+get_sms_length_cb (TpChannel *proxy,
+    guint chunks_required,
+    gint remaining_characters,
+    gint estimated_cost,
+    const GError *error,
+    gpointer user_data,
+    GObject *weak_object)
+{
+  GSimpleAsyncResult *result = user_data;
+  GetSmsLengthReturn *r;
+
+  if (error != NULL)
+    {
+      DEBUG ("Failed to get SMS length: %s", error->message);
+
+      g_simple_async_result_set_from_error (result, error);
+      goto out;
+    }
+
+  r = get_sms_length_return_new (chunks_required, remaining_characters,
+      estimated_cost);
+
+  g_simple_async_result_set_op_res_gpointer (result, r,
+      (GDestroyNotify) get_sms_length_return_free);
+
+out:
+  g_simple_async_result_complete (result);
+}
+
+/**
+ * tp_text_channel_get_sms_length_async:
+ * @self: a #TpTextChannel
+ * @message: a #TpClientMessage
+ * @callback: a callback to call when the request has been satisfied
+ * @user_data: data to pass to @callback
+ *
+ * Starts an async call to get the number of 140 octet chunks required to
+ * send a #message via SMS on #self, as well as the number of remaining
+ * characters available in the final chunk and, if possible,
+ * an estimate of the cost.
+ *
+ * Once the request has been satisfied, @callback will be called.
+ * You can then call tp_text_channel_get_sms_length_finish() to get the
+ * result of the operation.
+ *
+ * Since: 0.15.1
+ */
+void
+tp_text_channel_get_sms_length_async (TpTextChannel *self,
+    TpMessage *message,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GSimpleAsyncResult *result;
+
+  result = g_simple_async_result_new ((GObject *) self, callback, user_data,
+      tp_text_channel_get_sms_length_async);
+
+  tp_cli_channel_interface_sms_call_get_sms_length ((TpChannel *) self, -1,
+      message->parts, get_sms_length_cb, result, g_object_unref,
+      G_OBJECT (self));
+}
+
+
+/**
+ * tp_text_channel_get_sms_length_finish:
+ * @self: a #TpTextChannel
+ * @result: a #GAsyncResult
+ * @chunks_required: (out): if not %NULL used to return
+ * the number of 140 octet chunks required to send the message.
+ * @remaining_characters: (out): if not %NULL used to return
+ * the number of further characters that can be fit in the final chunk.
+ * A negative value indicates that the message will be truncated by
+ * abs(@remaining_characters).
+ * The value #G_MININT32 indicates the message will be truncated by
+ * an unknown amount.
+ * @estimated_cost: (out): if not %NULL used to return
+ * the estimated cost of sending this message.
+ * The currency and scale of this value are the same as the
+ * values of the #TpConnection:balance-scale and
+ * #TpConnection:balance-currency properties.
+ * A value of -1 indicates the cost could not be estimated.
+ * @error: a #GError to fill
+ *
+ * Finishes an async SMS length request.
+ *
+ * Returns: %TRUE if the number of 140 octet chunks required to send
+ * the message has been retrieved, %FALSE otherwise.
+ *
+ * Since: 0.15.1
+ */
+gboolean
+tp_text_channel_get_sms_length_finish (TpTextChannel *self,
+    GAsyncResult *result,
+    guint *chunks_required,
+    gint *remaining_characters,
+    gint *estimated_cost,
+    GError **error)
+{
+  GSimpleAsyncResult *simple = (GSimpleAsyncResult *) result;
+  GetSmsLengthReturn *r;
+
+  if (g_simple_async_result_propagate_error (simple, error))
+    return FALSE;
+
+  g_return_val_if_fail (g_simple_async_result_is_valid (result,
+        G_OBJECT (self), tp_text_channel_get_sms_length_async), FALSE);
+
+  r = g_simple_async_result_get_op_res_gpointer (simple);
+
+  if (chunks_required != NULL)
+    *chunks_required = r->chunks_required;
+
+  if (remaining_characters != NULL)
+    *remaining_characters = r->remaining_characters;
+
+  if (estimated_cost != NULL)
+    *estimated_cost = r->estimated_cost;
+
+  return TRUE;
+}
+
+/**
+ * tp_text_channel_ack_all_pending_messages_async:
+ * @self: a #TpTextChannel
+ * @callback: a callback to call when the messages have been acked
+ * @user_data: data to pass to @callback
+ *
+ * Acknowledge all the pending messages. This is equivalent of calling
+ * tp_text_channel_ack_messages_async() with the list of #TpSignalledMessage
+ * returned by tp_text_channel_get_pending_messages().
+ *
+ * Once the messages have been acked, @callback will be called.
+ * You can then call tp_text_channel_ack_all_pending_messages_finish() to get
+ * the result of the operation.
+ *
+ * See tp_text_channel_ack_message_async() about acknowledging messages.
+ *
+ * Since: 0.15.3
+ */
+void
+tp_text_channel_ack_all_pending_messages_async (TpTextChannel *self,
+    GAsyncReadyCallback callback,
+    gpointer user_data)
+{
+  GList *messages;
+
+  messages = g_queue_peek_head_link (self->priv->pending_messages);
+
+  tp_text_channel_ack_messages_async (self, messages,
+      callback, user_data);
+}
+
+/**
+ * tp_text_channel_ack_all_pending_messages_finish:
+ * @self: a #TpTextChannel
+ * @result: a #GAsyncResult
+ * @error: a #GError to fill
+ *
+ * Finish an asynchronous acknowledgement operation of all messages.
+ *
+ * Returns: %TRUE if the messages have been acked, %FALSE otherwise.
+ *
+ * Since: 0.15.3
+ */
+gboolean
+tp_text_channel_ack_all_pending_messages_finish (TpTextChannel *self,
+    GAsyncResult *result,
+    GError **error)
+{
+  return tp_text_channel_ack_messages_finish (self, result, error);
+}
