@@ -1085,12 +1085,10 @@ int
 main (int argc,
       char **argv)
 {
-  tp_tests_abort_after (10);
-  g_set_prgname ("call-example");
-  g_test_init (&argc, &argv, NULL);
+  tp_tests_init (&argc, &argv);
   g_test_bug_base ("http://bugs.freedesktop.org/show_bug.cgi?id=");
+  g_set_prgname ("call-example");
 
-  g_type_init ();
   future_cli_init ();
 
   g_test_add ("/call/basics", Test, NULL, setup, test_basics, teardown);
