@@ -1050,11 +1050,7 @@ int
 main (int argc,
       char **argv)
 {
-  tp_tests_abort_after (10);
-  g_type_init ();
-  tp_debug_set_flags ("all");
-
-  g_test_init (&argc, &argv, NULL);
+  tp_tests_init (&argc, &argv);
   g_test_bug_base ("http://bugs.freedesktop.org/show_bug.cgi?id=");
 
   have_ipv6 = check_ipv6_support ();
