@@ -23,11 +23,14 @@ typedef struct {
     /* Service side objects */
     TpBaseConnection *base_connection;
     ExampleEcho2Channel *chan_service;
+    ExampleEcho2Channel *sms_chan_service;
     TpHandleRepoIface *contact_repo;
+    TpHandle bob;
 
     /* Client side objects */
     TpConnection *connection;
     TpTextChannel *channel;
+    TpTextChannel *sms_channel;
 
     TpMessage *received_msg;
     TpMessage *removed_msg;
@@ -44,10 +47,10 @@ static void
 create_contact_chan (Test *test)
 {
   gchar *chan_path;
-  TpHandle handle, alf_handle;
   GHashTable *props;
 
   tp_clear_object (&test->chan_service);
+  tp_clear_object (&test->sms_chan_service);
 
   /* Create service-side tube channel object */
   chan_path = g_strdup_printf ("%s/Channel",
@@ -57,18 +60,14 @@ create_contact_chan (Test *test)
       TP_HANDLE_TYPE_CONTACT);
   g_assert (test->contact_repo != NULL);
 
-  handle = tp_handle_ensure (test->contact_repo, "bob", NULL, &test->error);
-
-  g_assert_no_error (test->error);
-
-  alf_handle = tp_handle_ensure (test->contact_repo, "alf", NULL, &test->error);
-  g_assert (alf_handle);
+  test->bob = tp_handle_ensure (test->contact_repo, "bob", NULL, &test->error);
   g_assert_no_error (test->error);
+  g_assert (test->bob != 0);
 
   test->chan_service = g_object_new (
       EXAMPLE_TYPE_ECHO_2_CHANNEL,
       "connection", test->base_connection,
-      "handle", handle,
+      "handle", test->bob,
       "object-path", chan_path,
       NULL);
 
@@ -81,8 +80,29 @@ create_contact_chan (Test *test)
   g_assert_no_error (test->error);
 
   g_free (chan_path);
+  g_hash_table_unref (props);
 
-  tp_handle_unref (test->contact_repo, handle);
+  /* Register channel implementing SMS */
+  chan_path = g_strdup_printf ("%s/ChannelSMS",
+      tp_proxy_get_object_path (test->connection));
+
+  test->sms_chan_service = g_object_new (
+      EXAMPLE_TYPE_ECHO_2_CHANNEL,
+      "connection", test->base_connection,
+      "handle", test->bob,
+      "object-path", chan_path,
+      "sms", TRUE,
+      NULL);
+
+  g_object_get (test->chan_service,
+      "channel-properties", &props,
+      NULL);
+
+  test->sms_channel = tp_text_channel_new (test->connection, chan_path,
+      props, &test->error);
+  g_assert_no_error (test->error);
+
+  g_free (chan_path);
   g_hash_table_unref (props);
 }
 
@@ -113,6 +133,7 @@ teardown (Test *test,
   test->mainloop = NULL;
 
   tp_clear_object (&test->chan_service);
+  tp_clear_object (&test->sms_chan_service);
 
   tp_cli_connection_run_disconnect (test->connection, -1, &test->error, NULL);
   g_assert_no_error (test->error);
@@ -127,6 +148,7 @@ teardown (Test *test,
   tp_clear_pointer (&test->sent_token, g_free);
 
   tp_clear_object (&test->channel);
+  tp_clear_object (&test->sms_channel);
 }
 
 static void
@@ -600,15 +622,224 @@ test_message_sent (Test *test,
   g_assert (test->sent_token == NULL);
 }
 
+static void
+notify_cb (GObject *object,
+    GParamSpec *spec,
+    Test *test)
+{
+  test->wait--;
+  if (test->wait <= 0)
+    g_main_loop_quit (test->mainloop);
+}
+
+
+static void
+test_sms_feature (Test *test,
+    gconstpointer data G_GNUC_UNUSED)
+{
+  gboolean is_sms;
+  GQuark features[] = { TP_TEXT_CHANNEL_FEATURE_SMS, 0 };
+
+  g_assert (tp_text_channel_get_sms_flash (test->sms_channel));
+
+  /* SMS feature is not prepared yet */
+  g_assert (!tp_text_channel_is_sms_channel (test->sms_channel));
+
+  g_object_get (test->sms_channel, "is-sms-channel", &is_sms, NULL);
+  g_assert (!is_sms);
+
+  test->wait++;
+  tp_proxy_prepare_async (test->sms_channel, features,
+      proxy_prepare_cb, test);
+
+  test->wait++;
+  g_signal_connect (test->sms_channel, "notify::is-sms-channel",
+      G_CALLBACK (notify_cb), test);
+
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  /* Feature has been prepared */
+  g_assert (tp_text_channel_is_sms_channel (test->sms_channel));
+
+  g_object_get (test->sms_channel, "is-sms-channel", &is_sms, NULL);
+  g_assert (is_sms);
+
+  /* Property is changed */
+  example_echo_2_channel_set_sms (test->sms_chan_service, FALSE);
+
+  test->wait++;
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  g_assert (!tp_text_channel_is_sms_channel (test->sms_channel));
+
+  g_object_get (test->sms_channel, "is-sms-channel", &is_sms, NULL);
+  g_assert (!is_sms);
+}
+
+#define MSG "Oh hi!"
+
+static void
+get_sms_length_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  Test *test = user_data;
+  guint chunks_required;
+  gint remaining_characters;
+  gint estimated_cost;
+
+  tp_text_channel_get_sms_length_finish (TP_TEXT_CHANNEL (source), result,
+      &chunks_required, &remaining_characters, &estimated_cost, &test->error);
+
+  g_assert_cmpuint (chunks_required, ==, strlen (MSG));
+  g_assert_cmpint (remaining_characters, ==,
+      EXAMPLE_ECHO_2_CHANNEL_MAX_SMS_LENGTH - strlen (MSG));
+  g_assert_cmpint (estimated_cost, ==, -1);
+
+  test->wait--;
+  if (test->wait <= 0)
+    g_main_loop_quit (test->mainloop);
+}
+
+
+static void
+test_get_sms_length (Test *test,
+    gconstpointer data G_GNUC_UNUSED)
+{
+  TpMessage *msg;
+
+  msg = tp_client_message_new_text (TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL, MSG);
+
+  tp_text_channel_get_sms_length_async (test->channel, msg,
+      get_sms_length_cb, test);
+
+  test->wait++;
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  g_object_unref (msg);
+}
+
+static void
+all_pending_messages_acked_cb (GObject *source,
+    GAsyncResult *result,
+    gpointer user_data)
+{
+  Test *test = user_data;
+
+  tp_text_channel_ack_all_pending_messages_finish (TP_TEXT_CHANNEL (source),
+      result, &test->error);
+
+  test->wait--;
+  if (test->wait <= 0)
+    g_main_loop_quit (test->mainloop);
+}
+
+static void
+test_ack_all_pending_messages (Test *test,
+    gconstpointer data G_GNUC_UNUSED)
+{
+  GQuark features[] = { TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, 0 };
+  GList *messages;
+  TpMessage *msg;
+
+  /* Send a first message */
+  msg = tp_client_message_new_text (TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL,
+      "Badger");
+
+  tp_text_channel_send_message_async (test->channel, msg, 0,
+      send_message_cb, test);
+
+  g_object_unref (msg);
+
+  /* Send a second message */
+  msg = tp_client_message_new_text (TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL,
+      "Snake");
+
+  tp_text_channel_send_message_async (test->channel, msg, 0,
+      send_message_cb, test);
+
+  g_object_unref (msg);
+
+  test->wait = 2;
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  tp_proxy_prepare_async (test->channel, features,
+      proxy_prepare_cb, test);
+
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  messages = tp_text_channel_get_pending_messages (test->channel);
+  g_assert_cmpuint (g_list_length (messages), ==, 2);
+
+  tp_text_channel_ack_all_pending_messages_async (test->channel,
+      all_pending_messages_acked_cb, test);
+
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  g_list_free (messages);
+
+  /* Messages have been acked so there is no pending messages */
+  messages = tp_text_channel_get_pending_messages (test->channel);
+  g_assert_cmpuint (g_list_length (messages), ==, 0);
+}
+
+static void
+test_pending_messages_with_no_sender_id (Test *test,
+    gconstpointer data G_GNUC_UNUSED)
+{
+  GQuark features[] = { TP_TEXT_CHANNEL_FEATURE_INCOMING_MESSAGES, 0 };
+  TpMessage *cm_message;
+  TpMessage *signalled_message;
+  GList *messages;
+  TpContact *sender;
+  gchar *text;
+
+  g_test_bug ("39172");
+
+  /* Deliberately passing sender=0 so we can set message-sender manually; if we set
+   * it here, or using tp_cm_message_set_sender(), message-sender-id will be
+   * filled in, which is exactly what we don't want.
+   */
+  cm_message = tp_cm_message_new_text (test->base_connection, 0,
+      TP_CHANNEL_TEXT_MESSAGE_TYPE_NORMAL, "hi mum");
+  tp_message_set_uint32 (cm_message, 0, "message-sender", test->bob);
+  g_assert_cmpstr (NULL, ==,
+      tp_asv_get_string (tp_message_peek (cm_message, 0), "message-sender-id"));
+  tp_message_mixin_take_received (G_OBJECT (test->chan_service), cm_message);
+
+  test->wait = 1;
+  tp_proxy_prepare_async (test->channel, features,
+      proxy_prepare_cb, test);
+  g_main_loop_run (test->mainloop);
+  g_assert_no_error (test->error);
+
+  messages = tp_text_channel_get_pending_messages (test->channel);
+  g_assert (messages != NULL);
+  g_assert_cmpuint (g_list_length (messages), ==, 1);
+
+  signalled_message = messages->data;
+  sender = tp_signalled_message_get_sender (signalled_message);
+  g_assert (sender != NULL);
+  g_assert_cmpstr (tp_contact_get_identifier (sender), ==, "bob");
+
+  text = tp_message_to_text ((TpMessage *) signalled_message, NULL);
+  g_assert_cmpstr (text, ==, "hi mum");
+  g_free (text);
+
+  g_list_free (messages);
+}
+
 int
 main (int argc,
       char **argv)
 {
-  tp_tests_abort_after (10);
-  g_type_init ();
-  tp_debug_set_flags ("all");
-
-  g_test_init (&argc, &argv, NULL);
+  tp_tests_init (&argc, &argv);
   g_test_bug_base ("http://bugs.freedesktop.org/show_bug.cgi?id=");
 
   g_test_add ("/text-channel/creation", Test, NULL, setup,
@@ -625,6 +856,15 @@ main (int argc,
       test_ack_message, teardown);
   g_test_add ("/text-channel/message-sent", Test, NULL, setup,
       test_message_sent, teardown);
+  g_test_add ("/text-channel/sms-feature", Test, NULL, setup,
+      test_sms_feature, teardown);
+  g_test_add ("/text-channel/get-sms-length", Test, NULL, setup,
+      test_get_sms_length, teardown);
+  g_test_add ("/text-channel/ack-all-pending-messages", Test, NULL, setup,
+      test_ack_all_pending_messages, teardown);
+
+  g_test_add ("/text-channel/pending-messages-with-no-sender-id", Test, NULL,
+      setup, test_pending_messages_with_no_sender_id, teardown);
 
   return g_test_run ();
 }
