@@ -11,6 +11,8 @@
 
 #include "bug-19101-conn.h"
 
+#include <telepathy-glib/interfaces.h>
+
 #include "debug.h"
 
 static void contacts_iface_init (gpointer g_iface, gpointer iface_data);
@@ -32,7 +34,7 @@ tp_tests_bug19101_connection_class_init (TpTestsBug19101ConnectionClass *klass)
 }
 
 /* A broken implementation of GetContactAttributes, which returns an empty dict
- * of attributes for each handle.
+ * of attributes for each handle (other than the self-handle).
  */
 static void
 tp_tests_bug19101_connection_get_contact_attributes (
@@ -42,12 +44,30 @@ tp_tests_bug19101_connection_get_contact_attributes (
     gboolean hold,
     DBusGMethodInvocation *context)
 {
-  GHashTable *result = g_hash_table_new_full (NULL, NULL, NULL,
-      (GDestroyNotify) g_hash_table_destroy);
+  TpBaseConnection *base_conn = TP_BASE_CONNECTION (iface);
+  GHashTable *result;
   guint i;
+  const gchar *assumed_interfaces[] = {
+    TP_IFACE_CONNECTION,
+    NULL
+  };
+
+  if (handles->len == 1 &&
+      g_array_index (handles, TpHandle, 0) == base_conn->self_handle)
+    {
+      DEBUG ("called for self-handle (during preparation), not being rubbish");
+      /* strictly speaking this should hold the handles on behalf of the
+       * sending process, but handles are immortal now anyway... */
+      result = tp_contacts_mixin_get_contact_attributes ((GObject *) iface,
+          handles, interfaces, assumed_interfaces, NULL);
+      goto finally;
+    }
 
   DEBUG ("called; returning rubbish");
 
+  result = g_hash_table_new_full (NULL, NULL, NULL,
+      (GDestroyNotify) g_hash_table_destroy);
+
   for (i = 0 ; i < handles->len ; i++)
     {
       TpHandle h= g_array_index (handles, TpHandle, i);
@@ -56,6 +76,7 @@ tp_tests_bug19101_connection_get_contact_attributes (
       g_hash_table_insert (result, GUINT_TO_POINTER(h), attr_hash);
     }
 
+finally:
   tp_svc_connection_interface_contacts_return_from_get_contact_attributes (
       context, result);
   g_hash_table_unref (result);
